/****************************************************************************
 *
 * Pocketop Wireless Keyboard(TM) add-on driver for Xwindow v1.0
 * Copyright GPL Licence. 
 * 2005.06.4 lampo1 <lampo1@paran.com>
 *
 ***************************************************************************/

#define MAXREPEAT 255

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#define MAXNAME 16 
#define SHIFT 50    /* most linux systems assign keycode 50 to Shift_L */

/* One IR key can have multiple meaning by combinations */
/* with other modifier keys. */

struct keymapping {
    unsigned char irbyte;

    char nml[MAXNAME];
    unsigned int kcode_nml;

    char punct[MAXNAME];
    unsigned int kcode_punct[2];

    char num[MAXNAME];
    unsigned int kcode_num[2];

    char func[MAXNAME];
    unsigned int kcode_func[2];
};  

struct keymapping KeyTable[54] = {
    {0x69,        "Tab",0,             "",{0,0},         "",{0,0}, "Escape",{0,0}},
    {0x01,          "q",0,       "exclam",{0,0},        "1",{0,0},       "",{0,0}},
    {0x09,          "w",0,           "at",{0,0},        "2",{0,0},       "",{0,0}},
    {0x11,          "e",0,   "numbersign",{0,0},        "3",{0,0},       "",{0,0}},
    {0x19,          "r",0,       "dollar",{0,0},        "4",{0,0},       "",{0,0}},
    {0x21,          "t",0,      "percent",{0,0},        "5",{0,0},       "",{0,0}},
    {0x29,          "y",0,  "asciicircum",{0,0},        "6",{0,0},       "",{0,0}},
    {0x31,          "u",0,    "ampersand",{0,0},        "7",{0,0},       "",{0,0}},
    {0x39,          "i",0,     "asterisk",{0,0},        "8",{0,0},       "",{0,0}},
    {0x41,          "o",0,    "parenleft",{0,0},        "9",{0,0},       "",{0,0}},
    {0x49,          "p",0,   "parenright",{0,0},        "0",{0,0},       "",{0,0}},
    {0x51,      "minus",0,         "plus",{0,0},    "slash",{0,0},       "",{0,0}},
    {0x59,  "BackSpace",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x77,  "Caps_Lock",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x07,          "a",0,        "grave",{0,0},         "",{0,0},       "",{0,0}},
    {0x0f,          "s",0,   "asciitilde",{0,0},         "",{0,0},       "",{0,0}},
    {0x17,          "d",0,        "slash",{0,0},         "",{0,0},       "",{0,0}},
    {0x1f,          "f",0,     "question",{0,0},         "",{0,0},       "",{0,0}},
    {0x27,          "g",0,  "bracketleft",{0,0},         "",{0,0},       "",{0,0}},
    {0x2f,          "h",0, "bracketright",{0,0},         "",{0,0},       "",{0,0}},
    {0x32,          "j",0,    "braceleft",{0,0},        "4",{0,0},       "",{0,0}},
    {0x3a,          "k",0,   "braceright",{0,0},        "5",{0,0},       "",{0,0}},
    {0x42,          "l",0,    "backslash",{0,0},        "6",{0,0},       "",{0,0}},
    {0x4a,  "semicolon",0,          "bar",{0,0}, "asterisk",{0,0},       "",{0,0}},
    {0x52, "apostrophe",0,        "equal",{0,0},         "",{0,0},       "",{0,0}},
    {0x5a,     "Return",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x65,    "Shift_L",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x03,          "z",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x0b,          "x",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x13,          "c",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x1b,          "v",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x23,          "b",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x33,          "n",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x3b,          "m",0,             "",{0,0},        "1",{0,0},       "",{0,0}},
    {0x43,      "comma",0,             "",{0,0},        "2",{0,0},       "",{0,0}},
    {0x4b,     "period",0,             "",{0,0},        "3",{0,0},       "",{0,0}},
    {0x53,         "Up",0,             "",{0,0},    "minus",{0,0},  "Prior",{0,0}},
    {0x7d,    "Shift_R",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x6d,  "Control_L",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x66,      "Alt_L",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x2b,      "space",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x34,      "space",0,             "",{0,0},         "",{0,0},       "",{0,0}},
    {0x3c,           "",0,             "",{0,0},        "0",{0,0},       "",{0,0}}, /* func */
    {0x44,       "Left",0,             "",{0,0},   "period",{0,0},   "Home",{0,0}},
    {0x4c,       "Down",0,             "",{0,0},     "plus",{0,0},   "Next",{0,0}},
    {0x54,      "Right",0,             "",{0,0},    "equal",{0,0},    "End",{0,0}},
    {0x5b,     "Delete",0,             "",{0,0},         "",{0,0},       "",{0,0}}, /* 47th */
    {0x40,           "",0,             "",{0,0},         "",{0,0},       "",{0,0}}, /* Date */
    {0x48,           "",0,             "",{0,0},         "",{0,0},       "",{0,0}}, /* Address */
    {0x50,           "",0,             "",{0,0},         "",{0,0},       "",{0,0}}, /* ToDo */
    {0x58,           "",0,             "",{0,0},         "",{0,0},       "",{0,0}}, /* Memo */
    {0x75,           "",0,             "",{0,0},         "",{0,0},       "",{0,0}}, /* Cmd */
    {0x76,           "",0,             "",{0,0},         "",{0,0},       "",{0,0}}, /* Punct */
    {0x6e,           "",0,             "",{0,0},         "",{0,0},       "",{0,0}}  /* Num */
};


int fd;
unsigned int idx;
unsigned char irbyte = 128;           /* initial state is break */
unsigned char prev_key;                     
unsigned int nrepeat = 0;
unsigned int num_FLAG = 0;
unsigned int punct_FLAG = 0;
unsigned int func_FLAG = 0;
unsigned int typematic_delay = 1000;   /* milliseconds */
unsigned int typematic_rate = 100;     /* milliseconds */
unsigned int auto_off_time = 300;      /* seconds */    

Display *d;

struct itimerval auto_off;
struct itimerval typematic;

void read_keymap() {
    FILE *fp;
    int i;
    int len = 254;
    char buf[255];
    char cset[] = " \n";
    char *p0, *p1, *p2;
 
    fp = popen("xmodmap -pke", "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot execute xmodmap -pke\n");
        exit(EXIT_FAILURE);
    }

    while ( fgets(buf, len, fp) ) {

	/* parsing one line of the result of xmodmap -pke */
	strtok(buf, cset);          /* discard "keycode" */
	p0 = strtok(NULL, cset);    /* point to keycode */
	strtok(NULL, cset);         /* discard "="  */
	p1 = strtok(NULL, cset);    /* point to X keysym name */ 
	p2 = strtok(NULL, cset);    /* point to the shifted X keysym name */

	for ( i = 0 ; i <= 46; i++) {

	    if ( (p1 != NULL) && (!strcmp( KeyTable[i].nml, p1)) ) {
		KeyTable[i].kcode_nml = atoi(p0);
	    }
	    
	    if ( (p1 != NULL) && (!strcmp( KeyTable[i].punct, p1)) ) {
		KeyTable[i].kcode_punct[0] = atoi(p0);
		KeyTable[i].kcode_punct[1] = 0;
	    }
	    
	    if ( (p2 != NULL) && (!strcmp( KeyTable[i].punct, p2)) ) {
		KeyTable[i].kcode_punct[0] = SHIFT;
		KeyTable[i].kcode_punct[1] = atoi(p0);
	    }

	    if ( (p1 != NULL) && (!strcmp( KeyTable[i].num, p1)) ) {
		KeyTable[i].kcode_num[0] = atoi(p0);
		KeyTable[i].kcode_num[1] = 0;
	    }
	    
	    if ( (p2 != NULL) && (!strcmp( KeyTable[i].num, p2)) ) {
		KeyTable[i].kcode_num[0] = SHIFT;
		KeyTable[i].kcode_num[1] = atoi(p0);
	    }

	    if ( (p1 != NULL) && (!strcmp( KeyTable[i].func, p1)) ) {
		KeyTable[i].kcode_func[0] = atoi(p0);
		KeyTable[i].kcode_func[1] = 0;
	    }
	    
	    if ( (p2 != NULL) && (!strcmp( KeyTable[i].func, p2)) ) {
		KeyTable[i].kcode_func[0] = SHIFT;
		KeyTable[i].kcode_func[1] = atoi(p0);
	    }
	}
    } 
    pclose(fp);

    #ifdef DEBUG
    printf("In printing...\n");
    for ( i = 0; i <= 53; i++ ) {
	printf("#%d %x %s %d %s %d %d %s %d %d %s %d %d\n",
	       i,
	       KeyTable[i].irbyte,
	       KeyTable[i].nml,
	       KeyTable[i].kcode_nml,
	       KeyTable[i].punct,
	       KeyTable[i].kcode_punct[0],
	       KeyTable[i].kcode_punct[1],
	       KeyTable[i].num,
	       KeyTable[i].kcode_num[0],
	       KeyTable[i].kcode_num[1],
	       KeyTable[i].func,
	       KeyTable[i].kcode_func[0],
	       KeyTable[i].kcode_func[1]
	       );
    }
    #endif
	       
}

void read_rcfile(FILE *fp) {

    char dmy[255];
    while ( (fgets(dmy, sizeof(dmy), fp)) != NULL ) {

	if ( dmy[0] == '#' ) continue;
	else if ( dmy[0] == 'd' ) {
	    sscanf(dmy, "%*s %d", &typematic_delay);
	} else if ( dmy[0] == 'r' ) {
	    sscanf(dmy, "%*s %d", &typematic_rate);
	} else if ( dmy[0] == 'a' ) {
	    sscanf(dmy, "%*s %d", &auto_off_time);
	}
    }
    fclose(fp);
}


void send_kcodes( unsigned int *ptr, unsigned char msb) {

    int j;

    if ( msb == 0 ) {       /* send kcodes in proper order */
	for ( j = 0; j <= 1; j++) {
	    if ( *ptr != 0 ) {
		XTestFakeKeyEvent(d, *ptr, (1^msb), 0);
		XFlush(d);

		#ifdef DEBUG
		printf("code= %d\n", *ptr);
		#endif
	    } 
	    ptr++;
	} 	
    } else {                /* send kcodes in reverse order */
	ptr++;
	for ( j = 0; j <= 1; j++) {
	    if ( *ptr != 0 ) {
		XTestFakeKeyEvent(d, *ptr, (1^msb), 0);
		XFlush(d);

		#ifdef DEBUG
		printf("code= %d\n", *ptr);
		#endif
	    } 
	    ptr--;
	} 	
    }
}

void translate( unsigned int i, unsigned char ch ) {
    
    unsigned char msb;   /* 0 : key pressed  */ 
			 /* 1 : key released */

    msb = (ch & 128) >> 7;   
    ch &= 127;  

    if ( ch == 0x6e ) {         /* Num key */  
	num_FLAG = (1^msb);
    } else if ( ch == 0x76 ) {  /* Punctuation Key */
	punct_FLAG = (1^msb);
    } else if ( ch == 0x3c ) {  /* function Key */
	func_FLAG = (1^msb);
    } else if ( num_FLAG == 1 ) {
	send_kcodes(&KeyTable[i].kcode_num[0], msb);		    
    } else if ( punct_FLAG == 1 ) {
	send_kcodes(&KeyTable[i].kcode_punct[0], msb);		    
    } else if ( func_FLAG == 1 ) {
	if ( ch == 0x59 ) {              /* pocketop off */
	    close(fd);
	    exit(EXIT_SUCCESS);
	}
	if ( (ch == 0x52) || (ch == 0xd2) ) {       /* keymap reload */          
	    read_keymap();                       /* func-apostrophe */
	    return;
	}
	send_kcodes(&KeyTable[i].kcode_func[0], msb);		    
    } else {
	XTestFakeKeyEvent(d, KeyTable[i].kcode_nml, (1^msb), 0);
	XFlush(d);

	#ifdef DEBUG
	printf("code= %d\n", KeyTable[i].kcode_nml);
	#endif
    }
}

void process_irbyte( unsigned int i, unsigned char ch ) {

    /* ignore duplicated break code of pocketop */
    if ( ch == prev_key ) return;       

    translate( i, ch );
    if ( (ch & 128) == 0) {
	setitimer(ITIMER_REAL, &typematic, NULL);
    } else {
	setitimer(ITIMER_REAL, &auto_off, NULL);
    }
    prev_key = ch;	
    return;
}

void alarm_inthandler() {
    if ( (128 & irbyte) == 0 ) {     /* if key is make, then repeat */
	nrepeat++;
	translate( idx, irbyte );
	if ( nrepeat >= MAXREPEAT ) {   /* not good solution, forgive me */
	    nrepeat = 0;
	    process_irbyte( idx, (irbyte | 128) ); /* send break of the key */
	    close(fd);
	    exit(EXIT_FAILURE);
	} 
    } else {        /* key is break state, so auto off */
	close(fd);
	exit(EXIT_SUCCESS);
    }
}
	
int main(int argc, char *argv[]) {

    int i;
    int nread, option;
    FILE *fp;
    char *device = "/dev/ttyS1";
    char *rcfile = "/etc/pktrc";
    struct termios oldtio, newtio;
    struct sigaction newhandler;
    sigset_t blocked;

    while ( (option = getopt(argc, argv, "hd:c:")) != -1 ) {
        switch(option) {
        case 'h' :
            fprintf(stderr,"Pocketop Wireless Keyboard driver 1.0\n");
            fprintf(stderr,"Usage : pktxwin [-d device] [-c rcfile]\n");
            fprintf(stderr,"example : pktxwin -d /dev/ttyS2 -c /etc/pktrc\n");
            exit(0);
        case 'd' :
            device = optarg;
            break;
        case 'c' :
            rcfile = optarg;
            break;
        default :
            fprintf(stderr,"unknown option.\n");
            exit(1);
        }
    }

    /* signal install */
    newhandler.sa_handler = alarm_inthandler;
    newhandler.sa_flags = SA_RESTART;
    sigemptyset(&blocked);
    sigaddset(&blocked,SIGALRM);

    newhandler.sa_mask = blocked;    
    if ( sigaction(SIGALRM, &newhandler, NULL) == -1 ) {
        perror("sigaction install error\n");
        exit(EXIT_FAILURE);
    }
    
    /* open irda port device */
    fprintf(stderr, "Opening %s\n", device);     
    fd = open(device, O_RDONLY | O_NOCTTY );
    if (fd == -1)  {
    	fprintf(stderr, "Cannot open %s\n", device);
    	exit(EXIT_FAILURE);
    }

    /* open rcfile */
    fprintf(stderr, "Opening %s\n", rcfile);
    fp = fopen( rcfile, "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot open %s\n", rcfile);
        fprintf(stderr,"Default time setting is applied.\n");
    } else {
	read_rcfile( fp );
    }

    /* read current xmodmap */
    read_keymap();

    /* Serial port setting to noncanonical, no echo, ...*/
    tcgetattr(fd, &oldtio);        
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;        
    newtio.c_lflag = 0;      
    newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
    newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 char received */       
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd,TCSANOW,&newtio);
               
    /* Timer setting */
    typematic.it_interval.tv_sec = typematic_rate / 1000;
    typematic.it_interval.tv_usec = (typematic_rate % 1000) * 1000;
    typematic.it_value.tv_sec = typematic_delay / 1000;
    typematic.it_value.tv_usec = (typematic_delay % 1000) * 1000; 

    auto_off.it_value.tv_sec = auto_off_time;
    auto_off.it_value.tv_usec = 0;
    

    d = XOpenDisplay(NULL);    
    setitimer(ITIMER_REAL, &auto_off, NULL);

    while (1) {
	nread = read(fd, &irbyte, 1);	  
	if ( nread > 0 ) {
	 
	    setitimer(ITIMER_REAL, &auto_off, NULL);
	    nrepeat = 0;

	    for ( i = 0; i <= 53; i++ ) {
		if ( (irbyte & 127) == KeyTable[i].irbyte ) {
		    idx = i;
		    process_irbyte(idx, irbyte);
		    break;
		}
	    }
	} else if ( nread == -1 ) {
	    fprintf(stderr,"IrDA Port reading error.\n");
	    fprintf(stderr,"Recommend removing other irda modules.\n");
	    exit(EXIT_FAILURE);
	}
    }  
}
